mod icf;

use std::fs::File;

use clap::{Parser, Subcommand};
use icf::serialize_icf;

use crate::icf::{
    crypto::{decrypt_icf, encrypt_icf, ICF_IV, ICF_KEY},
    decode_icf, fixup_icf,
    models::IcfData,
    parse_icf,
};

#[derive(Parser, Debug)]
struct Cli {
    #[command(subcommand)]
    command: Commands,
}

#[derive(Subcommand, Debug)]
enum Commands {
    #[command(about = "Fixes some common ICF errors, then encrypt the given ICF")]
    Encrypt { input: String, output: String },

    #[command(about = "Decrypts the given ICF")]
    Decrypt { input: String, output: String },

    #[command(about = "Decodes the given ICF (optionally to a JSON file)")]
    Decode {
        icf: String,
        json_output: Option<String>,
    },

    #[command(about = "Encodes a JSON file from the decode subcommand to an ICF")]
    Encode {
        json_input: String,
        output: String,
    },
}

fn main() -> Result<(), anyhow::Error> {
    let cli = Cli::parse();

    match &cli.command {
        Commands::Encrypt { input, output } => {
            let mut icf_buf = std::fs::read(input)?;
            fixup_icf(&mut icf_buf)?;
            let icf = parse_icf(&icf_buf)?;

            for entry in icf {
                println!("{}", entry.filename());
            }

            let encrypted_icf = encrypt_icf(&icf_buf, ICF_KEY, ICF_IV)?;

            std::fs::write(output, encrypted_icf)?;
        }
        Commands::Decrypt { input, output } => {
            let mut icf_buf = std::fs::read(input)?;
            let decrypted_icf = decrypt_icf(&mut icf_buf, ICF_KEY, ICF_IV)?;

            std::fs::write(output, decrypted_icf)?;
        }
        Commands::Decode { icf, json_output } => {
            let mut icf_buf = std::fs::read(icf)?;
            let icf = decode_icf(&mut icf_buf)?;

            if let Some(json_output) = json_output {
                let f = File::create(json_output)?;
                serde_json::to_writer_pretty(f, &icf)?;
            }

            for entry in icf {
                println!("{}", entry.filename())
            }
        }
        Commands::Encode { json_input, output } => {
            let input = std::fs::read_to_string(json_input)?;
            let icf: Vec<IcfData> = serde_json::from_str(&input)?;
            let out = serialize_icf(&icf)?;

            std::fs::write("test.bin", &out)?;

            let encrypted = encrypt_icf(&out, ICF_KEY, ICF_IV)?;

            std::fs::write(output, encrypted)?;
        }
    }

    Ok(())
}
